export PATH=/home/tiago/local/bin:$PATH
export PYTHONPATH=/home/tiago/workspace/SimuMotor:/home/tiago/workspace/SimuMotorCtrl:/home/tiago/workspace/SimuCounter:/home/tiago/workspace/SimuCoTiCtrl:$PYTHONPATH
